/*
    This Software (C) Copyright David Goodwin, 2007, 2008.
    
     This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
        

/*
 * ForceLTSView.java
 *
 * Created on 30 July 2007, 00:07
 */

package assignment;

import java.util.*;
import java.lang.*;
import javax.swing.*;
import java.awt.*;
import java.awt.geom.Point2D;
import javax.swing.Timer;
import java.awt.event.*;
import java.util.concurrent.*;

/**
 *
 * @author David
 */
public class ForceLTSView extends JPanel implements MouseListener,
                                                    MouseMotionListener{
    /** This represents a node in the graph. It holds position, velocity and
     * other such stuff.
     */
    protected class graphNode {
        public graphVector velocity;
        public graphVector position;
        
        public boolean selected;
        public boolean isStartNode;
        
        //public graphVector net_force;
        public LTSNode theNode;
        public ArrayList<graphNode> springs;
        public boolean noDraw;
        public graphNode() {
            velocity = new graphVector(0,0);
            position = new graphVector(0,0);
            //net_force = new graphVector(0,0);
            springs = new ArrayList<graphNode>();
            noDraw = false;
            selected = false;
        }
        public graphNode(LTSNode n) {
            velocity = new graphVector(0,0);
            position = new graphVector(0,0);
            //net_force = new graphVector(0,0);
            springs = new ArrayList<graphNode>();
            theNode = n;
            noDraw = false;
            selected = false;
        }
        public void randomizePosition() {
            position.x = rand.nextInt(getWidth());
            position.y = rand.nextInt(getHeight());
        }
    }
    
    protected LTS theLTS;
    protected int currentTool;
    
    protected Random rand;
    ArrayList<graphNode> theNodes;
    HashMap<String, graphNode> nodeMap = new HashMap<String,graphNode>();
    graphNode anchorNode;
    Semaphore moveMutex = new Semaphore(1);
    Semaphore flushMutex = new Semaphore(1);
    
    protected double repelconst = 8000;
    protected double springconst = 1.5;
    protected double timestep = 0.2;
    protected double damping = 0.45;
    
    javax.swing.Timer timer;
    protected int timerSpeed = 25;
    
    protected double total_kinetic_energy = 0;
    
    protected boolean showColors = false;
    HashMap<LTSNode, Integer> nodeColors = null;
    HashMap<Integer, Color> colorMap = null;
    int uniqueColors = 0;
    
    boolean showMessage = false;
    String messageTitle = "";
    String messageText = "";
    
    static int node_radius = 8;
    
    /** Creates a new instance of ForceLTSView */
    public ForceLTSView() {
        rand = new Random();
        theNodes = new ArrayList<graphNode>();
        addMouseListener(this);
        addMouseMotionListener(this);
        setBackground(Color.WHITE);
        repaint();
    }
    
    /** Toggle rendering mode. When enabled the renderer will run the layout
     * and rendering functions until the total kinetic energy drops below 50.
     */
    public void renderLTS() {
        //System.out.println("====== Rendering LTS ======");
        // Here we must begin the rendering process.
        
        /* The many steps to force-based success:
         * 1.1. Create nodes
         * 1.2. Set initial node velocities to (0,0)
         * 1.3. Set initial node positions randomly. No 2 nodes may be in the
         *      exact same place.
         */
        
        if (theLTS == null) return;
        
        if (timer != null) {
            if (timer.isRunning()){
                timer.stop();
                System.out.println("==== Render stopped ====");
                return;
            }
        }
        
        if (theLTS.getNodes().size() == 0) {
            System.out.println("Nothing to render.");
        }
        else {
            
            ActionListener al = new ActionListener(  ) {
                     public void actionPerformed(ActionEvent event) {
                         //System.out.println("======================================> Wow! the timer fired!");
                         total_kinetic_energy = 0;
                         try { moveMutex.acquire(); }
                    catch (Exception e) { e.printStackTrace(); } 
                         total_kinetic_energy = iterate();
                         
                         repaint();
                         moveMutex.release();
                         if (total_kinetic_energy < 50){
                             timer.stop();
                             System.out.println("===== Render Complete =====");
                         }
                     }
                };

            timer = new Timer(timerSpeed, al);
            timer.start(); 
            System.out.println("==== Render Running ====");
        }
    }
    
    /** Performs initial setup for the rendering and layout algorithms. This
     * should be called after changing the current LTS.
     */
    public void setup() {
        clearNodes();
        hideMessage();
        if (theLTS.getNodes().size() == 0) {
            System.out.println("Can't setup: no nodes.");
            showMessage("Unable to render: No LTS loaded", "There is currently no LTS loaded. Please load an LTS and refresh the display.");
        }
        else {
            createNodes();
            total_kinetic_energy=0;
        }
        // We need to stick the anchor node in the middle
        LTSNode gnn = new LTSNode();
        gnn.setName("$ANCHOR_NODE$$");
        anchorNode = new graphNode(gnn);
        anchorNode.position.x = getWidth() / 2;
        anchorNode.position.y = getHeight() / 2;
        anchorNode.noDraw = true;
        updateColors();
    }
    
    /** Creates graphNode wrappers for all nodes in the current LTS.
     */
    protected void createNodes() {
        System.out.println("Creating nodes...");
        Collection<LTSNode> nodes = theLTS.getNodes();
        
        // Now create the start node:
        if (theLTS.getStartNode() != null) {
            graphNode sn = createNode(theLTS.getStartNode());
            sn.isStartNode = true;
            storeNode(sn);
        }
       
        for(LTSNode n:nodes) {
            graphNode gn = createNode(n);
            storeNode(gn);
        }
        
        for (graphNode n:theNodes) {
            anchorNode.springs.add(n);
            n.springs.add(anchorNode);
        }
       
        
    }
    
    /** Handles the creation of a graphNode. If the node already exists it will
     * return the existing instance. If it does not exist it will create it.
     *
     * Node creation is recursive if there are any transitions coming out of the
     * node.
     */
    protected graphNode createNode(LTSNode n) {
        if (nodeMap.containsKey(n.toString())) {
            // The node already exists, return it
            return nodeMap.get(n.toString());
        }
        else {
            // The node doesnt exist, create it.
            graphNode gn = new graphNode(n);
            gn.randomizePosition();
            //System.out.println("   Node create: " + n.toString() + " at position " + gn.position.toString());
            Collection<LTSTransition> ltst = theLTS.getTransitions();
            storeNode(gn);
            for (LTSTransition l:ltst) {
                if (l.getNode1().equals(n)) {
                    if (l.getNode2() != null) {
                        if (l.getNode1().equals(l.getNode2())) {
                            // This is a loop node. We cant do a recursive node
                            // create on it (and we dont need to).
                            
                            // Infact, we dont even need a spring for loops
                            //gn.springs.add(gn);
                        }
                        else {
                            graphNode nnn = createNode(l.getNode2());
                            nnn.springs.add(gn);
                            gn.springs.add(nnn);
                            storeNode(nnn);
                        }
                    }
                }
            }
            return gn;
        }
    }
   
    /** Stores a node for future use.
     */
    protected void storeNode(graphNode n) {
        if (n == null) return;
        if (!theNodes.contains(n)) 
        {
            theNodes.add(n);
            nodeMap.put(n.theNode.toString(),n);
        }
    }
    
    /** Clears all graphNode wrappers used by the layout algorithm.
     */
    protected void clearNodes() {
        theNodes = new ArrayList<graphNode>();
        nodeMap = new HashMap<String,graphNode>();
    }
    
    /** Generates position data for the nodes. [[ DEPRICATED ]]
     */
    protected void runSim() {
        System.out.println("Running Simulation...");
         do {
             total_kinetic_energy = 0;
             
             total_kinetic_energy = iterate();
             repaint();

         } while (total_kinetic_energy > 50);
         // Above: Loop until total kinetic energy is less than some small number (the simulation has stopped running)

    }
    
    /** This performs one iteration of the spring embedder algorithm.
     */
    public double iterate() {
        double total_kinetic_energy = 0;
        graphVector net_force;
        for (graphNode n:theNodes) {
                 net_force = new graphVector(0,0);
                 
                 // For each other node...
                 for (graphNode nn:theNodes) {
                     if (nn != n) {
                         // net_force := net_force + Coulomb_repulsion( this_node, other_node )
                         net_force.add(Coulomb_repulsion(n,nn));
                     }
                 }
                 
                 // For each spring connected to this node...
                 for (graphNode spring:n.springs) {
                     // net_force := net_force + Hooke_attraction( this_node, spring )
                     net_force.add(Hooke_attraction(n,spring)); 
                 }
                 
                 // Without damping, it moves forever:
                 // this_node.velocity := (this_node.velocity + (timestep * net_force)) * damping
                 // timestep * net_force
                 graphVector vma = net_force.scalarMultiply(net_force,timestep);
                 // velocity + (timestep * net_force)
                 graphVector vmb = graphVector.add(n.velocity,vma);
                 // (this_node.velocity + (timestep * net_force)) * damping
                 vmb.scalarMultiply(damping);
                 n.velocity = vmb;
                 /*n.velocity = n.net_force.scalarMultiply(
                         n.net_force.add(n.velocity,n.net_force.scalarMultiply(n.net_force,timestep)),
                         damping);*/
                 
                 // this_node.position := this_node.position + (timestep * this_node.velocity)
                 
                 // We dont want to move this node if it is being dragged:
                 if (n != selectedNode) {
                    n.position.add(graphVector.scalarMultiply(n.velocity,timestep));
                 }
                 
                 
                 // total_kinetic_energy := total_kinetic_energy + (this_node.velocity)^2
                 total_kinetic_energy += Math.pow(n.velocity.length(),2);
             }
        
        repaint();
        return total_kinetic_energy;
    }
    
    /** Needed by runSim()
     */
    protected graphVector Coulomb_repulsion(graphNode a, graphNode b) {
        graphVector gva = new graphVector (a.position.x - b.position.x, a.position.y - b.position.y);
        double force = repelconst * (200) / Math.pow (gva.length(), 2);
        graphVector gvau = gva.getUnit();
        gvau.scalarMultiply(force);
        return gvau;
    }
    /** Needed by runSim()
     */
    protected graphVector Hooke_attraction(graphNode a, graphNode spring) {
        graphVector d = new graphVector ( a.position.x -spring.position.x,a.position.y - spring.position.y);//   
        d.x *= -springconst;
        d.y *= -springconst;

        return d;
    }
    
    /** Renders the calculated node graph thing
     */
    protected void renderGraph() {
        System.out.println("  Rendering...");
        paintComponent(this.getGraphics());
        
        for (graphNode n:theNodes) {
            System.out.println("  Node '" + n.theNode.toString() + "' is at position '" + n.position.toString() + "'");
        }
        
    }
    
    /** Renders the LTS and other stuff that should be on screen in the layout/editing/rendering widget
     */
    public void paintComponent(Graphics g) {
     
        super.paintComponent(g);
        
        if (theLTS == null) return;
        
        g.drawString("Total Kinetic Energy: " + total_kinetic_energy + "      Node Count: " + theNodes.size(),20,20);
        
        if (theNodes.size() != 0) {

            // Draw Transitions
            Collection<LTSTransition> transitions = theLTS.getTransitions();
            for (LTSTransition t:transitions) {
                if (t.getNode2() == null) continue;

                if (t.getNode2().equals(t.getNode1()))
                    drawLoop(g,createNode(t.getNode1()), t.getName());
                else
                    drawSpring(g,createNode(t.getNode1()),createNode(t.getNode2()),t.getName());
            }
            
            // Draw nodes
            for (graphNode n:theNodes) {
                drawNode(g,n);
            }

            // Lastly, if creating a new transition:
            if (tropMode) {
                if (currentTool == 4) {
                    // We are adding a transition
                    // Lets draw a transition line!
                    
                    graphNode g2 = new graphNode();
                    g2.position.x = tropPoint.x;
                    g2.position.y = tropPoint.y;
                   
                    graphVector vec = new graphVector(g2.position.x - tropNodeA.position.x, 
                            g2.position.y - tropNodeA.position.y).getUnit();

                    double dx = vec.x * node_radius;
                    double dy = vec.y * node_radius;

                    drawArrow (g, (int)(tropNodeA.position.x + node_radius + dx), 
                            (int)(tropNodeA.position.y + node_radius + dy),
                            (int)(g2.position.x), (int)(g2.position.y));

                    // Draw the transition label
                    Double xx = (tropNodeA.position.x + g2.position.x)/2;
                    Double yy = (tropNodeA.position.y + g2.position.y)/2;
                    Point p3 = new Point();
                    p3.setLocation(xx,yy);

                }
            }
        }
        else {
            System.out.println("LTSRender: nothing to render.");
        }
        
        // And now draw the message if there is one to draw.
        if (showMessage) {
            // Yep, there is a message to show. The message should be in the center
            // of the widget.
            
            // First draw the box. 400w, 125t in the center
            
            int hPos = this.getWidth() / 2;
            int vPos = this.getHeight() / 2;
            vPos -= 62.5;
            hPos -= 200;
            g.setColor(Color.WHITE);
            g.fillRect(hPos, vPos, 400, 125);
            g.setColor(Color.BLACK);
            g.drawRect(hPos, vPos, 400, 125);
            
            // Now draw the title bar:
            g.setColor(Color.BLUE);
            g.fillRect(hPos, vPos, 400, 19);
            g.setColor(Color.BLACK);
            g.drawRect(hPos, vPos, 400, 19);
            
            // Now draw the title text:
            Font theFont = g.getFont();
            g.setColor(Color.WHITE);
            g.setFont(theFont.deriveFont(Font.BOLD));
            g.drawString(messageTitle, hPos + 5, vPos + 15);
            g.setFont(theFont);
            
            // And now for the message text. This is a bit more complicated as
            // we have to wrap the text.
            g.setColor(Color.BLACK);
            if (messageText.length() <= 69)
                g.drawString(messageText ,hPos+15, vPos+35);
            else {
                // We have to split it every 45 characters unless we can find
                // a space within 10 characters in which case we will split there.
                String[] lines = wrapString(messageText, 65);
                hPos += 15;
                vPos += 35;
                for(String s:lines) {
                    if (s != null) {
                        g.drawString(s, hPos, vPos);
                        vPos += 15;
                    }
                }
            }
        }
    }

    /** Draws the specified node on the specified graphics object
     */
    protected void drawNode(Graphics g, graphNode gn) {
        
        if (gn.noDraw) return;
        
        Point p = new Point();
        p.setLocation((gn.position.x ),(gn.position.y ));

        // Set the selection colour. We should remember the existing one too.
        Color tmp = g.getColor();
        if (showColors) {
            // Get the node color:
            Integer nc = nodeColors.get(gn.theNode);
            
            if (colorMap.containsKey(nc)){
                // Set the color:
                g.setColor(colorMap.get(nc));
            }
        }
        else {
            if (gn.selected) {
                g.setColor(Color.red);
            }
            else if (gn == tropNodeA) {
                g.setColor(Color.green);
            }
            else{
                g.setColor(Color.blue);
            }
        }
        
        int diameter = node_radius * 2;

        // draw the node
        if (showColors && gn.selected) {
            // The node is selected. Because we are showing colors we cant simply
            // use the color to show it is selected. So we show the selection
            // by drawing a larger oval around the outside.
            Color old = g.getColor();
            g.setColor(Color.RED);
            g.fillOval(p.x - 3, p.y - 3, diameter+7, diameter+7);
            g.setColor(old);
        }
        if (showColors && (gn == tropNodeA)) {
            // The node is selected in a transition operation. We are showing
            // colors so we cant simply show it is selected. So we show the
            // selection by drawing a larger oval around the outside.
            Color old = g.getColor();
            g.setColor(Color.green);
            g.fillOval(p.x - 3, p.y - 3, diameter+7, diameter+7);
            g.setColor(old);
        }
        g.fillOval(p.x, p.y, diameter+1, diameter+1);
        g.setColor(tmp);
        g.drawOval(p.x, p.y, diameter, diameter);
        
        p.setLocation(gn.position.x + node_radius, gn.position.y + node_radius);
        g.drawString(gn.theNode.getNodeName(), p.x + node_radius , p.y + node_radius);
        
        if (gn.isStartNode) {
            // Its the start node.
            g.drawOval(p.x-(node_radius+3),p.y - (node_radius+3),diameter+6,diameter+6);
        }
    }
    
    /** Draws a transition loop (a transition like node1/label/node1)
     */
    protected void drawLoop(Graphics g, graphNode gn, String name) {        
        g.drawOval((int)gn.position.x - (node_radius - 2), (int)gn.position.y - (node_radius - 2), node_radius*2, node_radius*2);
        g.drawString(name, (int)gn.position.x - (node_radius - 2), (int)gn.position.y - (node_radius - 2));
    }
    
    /** Draws a transition from node g1 to node g2 with the specified name
     */
    protected void drawSpring(Graphics g, graphNode g1, graphNode g2, String name) {
        
        if (g1.noDraw || g2.noDraw) return;

        // Draw the transition:
        graphVector vec = new graphVector(g2.position.x - g1.position.x, g2.position.y - g1.position.y).getUnit();
    
        double dx = vec.x * node_radius;
        double dy = vec.y * node_radius;
    
        drawArrow (g, (int)(g1.position.x + node_radius + dx), (int)(g1.position.y + node_radius + dy),
	       (int)(g2.position.x + node_radius  - dx), (int)(g2.position.y + node_radius  - dy));

        // Draw the transition label
        Double xx = (g1.position.x + g2.position.x)/2;
        Double yy = (g1.position.y + g2.position.y)/2;
        Point p3 = new Point();
        p3.setLocation(xx,yy);
        g.drawString(name,p3.x ,p3.y);
    }
    
    /** Draws an arrow
     */
    protected void drawArrow(Graphics g, int x1, int y1, int x2, int y2) {
        Point p1 = new Point(x1, y1);
        Point p2 = new Point(x2, y2);
        drawArrow(g, p1, p2);
    }
    
    /** Also draws an arrow
     */
    protected void drawArrow(Graphics g, Point p1, Point p2) {
        // first draw the shaft of the arrow:

        int x1 = p1.x;
        int x2 = p2.x;
        int y1 = p1.y;
        int y2 = p2.y;
        
        int dx = x2 - x1;
        int dy = y2 - y1;

        double length = Math.sqrt (Math.pow (dx, 2) + Math.pow (dy, 2));

        double frac = 4.0 / length;

        g.drawLine (x1, y1, x2, y2);
        g.drawLine (x1 + (int)((1 - frac) * dx + frac * dy), y1 + (int)((1 - frac) * dy - frac * dx), x2, y2);
        g.drawLine (x1 + (int)((1 - frac) * dx - frac * dy), y1 + (int)((1 - frac) * dy + frac * dx), x2, y2);
    }
    
    /** Toggles the node color view. If the node color view has not been enabled
     * for this LTS yet it will run the coloring algorithm and setup the required
     * color maps.
     */
    public void showColors() {
        // Toggle the mode
        if (showColors) showColors = false;
        else showColors = true;
        
        if (nodeColors == null) {
            showMessage("Please wait...", "Colouring node graph.....");
            // If coloring is on, get the colors:
            if (showColors) {
                nodeColors = theLTS.colorLTS(theLTS);
            }

            // Calculate the number of unique colors:
            HashSet<Integer> uC = new HashSet<Integer>();
            for (LTSNode n:theLTS.getNodes()) {
                uC.add(nodeColors.get(n));
            }
            uniqueColors = uC.size();
            
            System.out.println("Unique colors: " + uniqueColors);
            
            colorMap = new HashMap<Integer, Color>();
            
            // Setup the color map:
            if (uniqueColors <= 13) {
                // Use the contants
                int nc = 0;
                for (Integer i:uC) {
                    Color nColor = null;
                    if (nc == 0)        nColor = Color.BLACK;
                    else if (nc==1)     nColor = Color.BLUE;
                    else if (nc==2)     nColor = Color.CYAN;
                    else if (nc==3)     nColor = Color.DARK_GRAY;
                    else if (nc==4)     nColor = Color.GRAY;
                    else if (nc==5)     nColor = Color.GREEN;
                    else if (nc==6)     nColor = Color.LIGHT_GRAY;
                    else if (nc==7)     nColor = Color.MAGENTA;
                    else if (nc==8)     nColor = Color.ORANGE;
                    else if (nc==9)     nColor = Color.PINK;
                    else if (nc==10)    nColor = Color.RED;
                    else if (nc==11)    nColor = Color.WHITE;
                    else if (nc==12)    nColor = Color.YELLOW;
                    colorMap.put(i,nColor);
                    System.out.println("Mapping node color " + i + " to RGB color " + nColor.getRGB());
                    nc++;
                }
            }
            else {
                // There is too many colors to map onto the constants. Generate
                // random colors
                for (Integer i:uC) {
                    Color nColor = new Color(rand.nextInt());
                    System.out.println("Mapping node color " + i + " to random RGB color " + nColor.getRGB());
                    colorMap.put(i,nColor);
                }
            }
            hideMessage();
        }
        
        // Redraw the component.
        repaint();
    }
    
    /** Recolors the graph
     */
    public void updateColors() {
        nodeColors = null;
        uniqueColors = 0;
        colorMap = null;
        if (showColors){
            showColors = false;
            showColors();
        }
    }
    
    /** Shows a message in the center of the rendering widget. Causes the widget
     * to repaint.
     * @param title The title of the message
     * @param text The text of the message
     */
    public void showMessage(String title, String text) {
        showMessage = true;
        messageText = text;
        messageTitle = title;
        repaint();
    }
    
    /** Hides the currently showing message. Causes the widget to repaint.
     */
    public void hideMessage() {
        showMessage = false;
        messageText = "";
        messageTitle = "";
        repaint();
    }
    
    /** A function used by the code for displaying a message in the rendering widget.
     * Attempts to wrap a string to a specified column count. It will try to insert
     * line breaks between words.
     * @param text the string to wrap
     * @param column The number of columns to try and wrap to
     * @return The wrapped string, one line per array item.
     */
    protected String[] wrapString(String text, int column) {
        ArrayList<String> lines = new ArrayList<String>();
        
        String[] words = text.split(" ");
        int currentLineLength = 0;
        StringBuilder sb = new StringBuilder();
        for (String s:words) {
            if ((currentLineLength + s.length() + 1) <= column){
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
            else if (currentLineLength > 0) {
                lines.add(sb.toString());
                sb = new StringBuilder();
                currentLineLength = 0;
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
            else {
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
        }
        if (!sb.toString().equals("")) {
            lines.add(sb.toString());
        }
        return lines.toArray(new String[1]);
    }
    
    
    /** Sets the LTS to display/edit
     * @param anLTS the LTS to display/edit
     */
    public void setLTS(LTS anLTS) {
        theLTS = anLTS;
    }
    
    /** Returns the current LTS
     * @returns The current LTS
     */
    public LTS getLTS() {
        return theLTS;
    }
    

    /** Returns the currentt tool
     * @returns The current tool
     */
    public int getTool() {
        return currentTool;
    }
    /**************************************************************************/
    /*                      Tool Variables                                    */
    /**************************************************************************/
    graphNode selectedNode;
    ArrayList<graphNode> spareNodes = new ArrayList<graphNode>();
    boolean dragMode;
    boolean tropMode;
    graphNode tropNodeA;
    Point tropPoint = new Point();    

    /**************************************************************************/
    /*                      Tool Functions                                    */
    /**************************************************************************/
    
    /** Sets the current tool for subclasses that support it.
     * @param tool The tool number to set. The meaning of the tool number depends
     *  on the class implementing the tool
     */
    public void setTool(int tool) {
        currentTool = tool;
        if (currentTool == 8) {
            // Deselect what ever is selected and switch to tool 0;
            currentTool = 0;
            deselectNode();
            repaint();
        }
    }
    
    /** Selects the specified node. Selected nodes are drawn with a different
     * colour.
     */
    public void selectNode(graphNode n) {
        if (n == null) return;
        
        deselectNode();
        
        n.selected = true;
        selectedNode = n;
        repaint();
    }
    /** Deselects the current node.
     */
    public void deselectNode() {
        if (selectedNode == null) return;
        selectedNode.selected = false;
        selectedNode = null;
    }
    /** Sets the start node to the one specified.
     */
    public void setStartNode(graphNode n) {
        
        if (n == null) return;
        
        graphNode sn = createNode(theLTS.getStartNode());
        
        sn.isStartNode = false;
        n.isStartNode = true;
        
        theLTS.setStartNode(n.theNode);
        
        repaint();
    }
    /** Adds a new node at the specified position. New nodes are added to a
     * list of spare nodes until they are put into a transition or flushChanges
     * is called.
     */
    public void addNode(int x, int y) {
        // We have the position, now get the name.
        String nodeName = JOptionPane.showInputDialog("Enter a name for this node:");
        LTSNode newNode = new LTSNode();
        newNode.setName(nodeName);
        graphNode gn = new graphNode(newNode);
        gn.position.x = x;
        gn.position.y = y;
        
        spareNodes.add(gn);
        
        anchorNode.springs.add(gn);
        gn.springs.add(anchorNode);
        
        storeNode(gn);
        
        selectNode(gn);
        updateColors();
        repaint();
    }
    /** Adds new transitions. The first call will specify the first node and
     * the second node will specify the second node, create the transition
     * and add it to the LTS. It will also tidy up the LTS data structures
     * to remove unneeded single-node placeholder transitions and other such
     * things.
     */
    public void addTransition(graphNode n) {
        if (n == null) return;
        if (!tropMode) {
            // We arent in transition operation mode yet. We must be in the
            // first stage.
            tropMode = true;
            tropNodeA = n;
            
        }
        else {
            
            String transitionName = JOptionPane.showInputDialog("Enter a name for this transition:");
            
            if (transitionName == null) {
                // The user pressed cancel.
                cancelTransitionOp();
                return;
            }
            // We are in second stage trop mode.
            tropMode = false;
            tropNodeA.springs.add(n);
            n.springs.add(tropNodeA);
            
            if (spareNodes.contains(tropNodeA))
                spareNodes.remove(tropNodeA);
            if (spareNodes.contains(n))
                spareNodes.remove(n);
            
            LTSTransition lt = new LTSTransition();
            lt.setNode1(tropNodeA.theNode);
            lt.setNode2(n.theNode);
            lt.setName(transitionName);
            theLTS.addTransition(lt);
            tropNodeA = null;
            updateColors();
            repaint();
        }
    }
    /** Returns true if we are currently in the middle of adding a new transition
     */
    public boolean inTransitionOpMode() {
        return tropMode;
    }
    /** Cancels the current transition operation (add or remove transition)
     */
    public void cancelTransitionOp() {
        tropMode = false;
        tropNodeA = null;
        repaint();
    }
    /** Removes the node under the mouse pointer and all transitions connected to it.
     */
    public void removeNode() {
        if (selectedNode != null) {
            // Remove the node
            graphNode n = selectedNode;
            deselectNode();
            
            // Remove all transitions linking to this node
            for (graphNode nn:n.springs) {
                nn.springs.remove(n);
            }
            
            // de-anchor it:
            anchorNode.springs.remove(n);
            
            // Remove any transitions involving this node from the LTS:
            for (LTSTransition t:theLTS.getTransitions()) {
                if (t.getNode1().equals(n.theNode) ||
                    t.getNode2().equals(n.theNode)) {
                    // This transition includes the node to be removed.
                    // We must remove the transition
                    theLTS.removeTransition(t);
                }
            }
            
            // Now we must check the start node reference:
            if (n == createNode(n.theNode))
                for (LTSTransition t:theLTS.getTransitions()) {
                    theLTS.setStartNode(t.getNode1());
                    createNode(theLTS.getStartNode()).isStartNode = true;
                    break;
                }
                
            // Ok, all transitions that this node takes part in should be 
            // removed from the local data and the LTS. If it was a start node
            // a new start node has been set. Time to finally remove the node.
            theNodes.remove(n);
            nodeMap.remove(n.theNode.toString());
            
            n.springs.clear();
            
            updateColors();
            
            repaint();
        }
    }
    /** This is like the addTransition function but it removes transitions instead
     */
    public void removeTransition(graphNode n) {
        if (n == null) return;
        if (!tropMode) {
            // We arent in transition operation mode yet. We must be in the
            // first stage.
            tropMode = true;
            tropNodeA = n;
            
        }
        else {
            // We are in second stage trop mode.
            tropMode = false;

            // Ok, we now have both nodes: tropNodeA and n
            
            
            // First up, shift the transition out of the LTS class
            
            ArrayList<LTSTransition> trans = new ArrayList<LTSTransition>();
            LTSTransition theTrans = null;
            for (LTSTransition t:theLTS.getTransitions())
                if (t.getNode1().equals(tropNodeA.theNode) & t.getNode2().equals(n.theNode))
                    trans.add(t);
            
            // Return if the transition does not exist.
            if (trans.isEmpty()) return;
            
            if (trans.size() > 1) {
                // We have multiple nodes. We will have to ask the user which
                // to remove.
                ArrayList<String> listOfTransitionNames = new ArrayList<String>();
                
                for (LTSTransition t:trans)
                    listOfTransitionNames.add(t.getName());
                
                Object[] options = listOfTransitionNames.toArray();
                String result = (String)JOptionPane.showInputDialog(null, 
                        "Multiple transitions were found between the two nodes " +
                        "you selected.\nWhich transition do you wish to remove?", 
                        "Remove Transition",
                        JOptionPane.QUESTION_MESSAGE, 
                        null, 
                        options, 
                        null);
                
                if (result == null) {
                    tropNodeA = null;
                    return;
                }
                
                for (LTSTransition t:trans)
                    if (t.getNode1().equals(tropNodeA.theNode) &&
                        t.getNode2().equals(n.theNode) &&
                        t.getName() == result) {
                        theLTS.removeTransition(t);
                    }
            }
            else {
                // there is only one transition - remove it.
                theLTS.removeTransition(trans.get(0));
            }
            
            tropNodeA = null;
            
            updateColors();
            
            repaint();
        }
    }
    /** Starts the drag timer and enables drag time.
     */
    public void startMove() {
        dragMode = true;
    }
    /** Stops the drag timer and disables drag mode.
     */
    public void stopMove() {
        dragMode = false;
        deselectNode();
    }
    /** This function gets the graphNode under the specified point. If there is
     * nothing under the point null is returned.
     * @param x The x position
     * @param y The y position
     * @return the node under the point if there is one, otherwise null.
     */
    public graphNode getNodeUnderPoint(int x, int y) {
        /* The node width/length is 10 (radius of 5). The position vector points
         * to the top left corner of the rectangle surrounding the circle.
         *
         * For the point to be over a node:
         *  node.x < x < node.x+10
         *  node.y < y < node.y+10
         */
        
        int diameter = node_radius * 2;
        
        System.out.println("Searching for node under (" + x + "," + y + ")");
        
        for (graphNode n:theNodes)
            if ((n.position.x <= x) && (x < n.position.x + diameter))
                if ((n.position.y <= y) && (y < n.position.y + diameter))
                    return n;
        
        System.out.println("Not found");
        // We couldnt find a node. Return null.
        return null;
    }
    
    /** This function flushes changes made here back to the LTS data strucutres
     */
    public void flushChanges() {
        try {
        flushMutex.acquire();
        } catch(Exception e) {e.printStackTrace();}
        /* Here we need to flush any changes made to the internal data
         * strucutres back to the LTS data strucutres.
         *
         * The following changes are possible:
         *  - Add node
         *  - Add transition
         *    - Could make use of a currently unallocated node
         *  - Remove node
         *  - Remove transition
         *    - Could make a currently allocated node unallocated
         *
         *  Transition changes are automaticaly synchronized - all we have to
         *  do is sync node changes. This includes:
         *    - Flush any unallocated nodes that are now part of a transition
         *    - Add any newly unallocated nodes
         */
        
        // get existing transitions/nodes:
        Collection<LTSTransition> trans = theLTS.getTransitions();
        ArrayList<LTSTransition> toRemove = new ArrayList<LTSTransition>();
        
        // We need to pull all lone nodes out of the transition list
        HashSet<LTSNode> loneNodes = new HashSet<LTSNode>();
        // populate the lone node list:
        for (LTSTransition t:trans) {
            if (t.getNode2() == null){
                loneNodes.add(t.getNode1());
                toRemove.add(t);
            }
        }
        for (LTSTransition t:toRemove)
            theLTS.removeTransition(t);
        
        
        // Ok, we now have two lists of nodes:
        //   -- spareNodes : These have not been allocated
        //   -- loneNodes  : These may have been allocated by a transition add
        //                   operation but not cleaned up.
        trans = theLTS.getTransitions();
        // first up, remove any newly allocated nodes that are in the "possibly unallocated"
        // list.
        // We must see if node n is part of any transition.
        ArrayList<LTSNode> toRemoven = new ArrayList<LTSNode>();
        for (LTSNode n:loneNodes)
            for (LTSTransition t:trans)
                if (t.getNode2() != null)
                    if (t.getNode1().equals(n) || t.getNode2().equals(n))
                        toRemoven.add(n);
        for (LTSNode n:toRemoven)
            loneNodes.remove(n);
        
        // Now we can shift the spare nodes:
        for (graphNode gn:spareNodes)
            loneNodes.add(gn.theNode);
        
        // And now we create the dummy transitions
        for (LTSNode n:loneNodes)
            theLTS.addTransition(new LTSTransition(n,null,""));
        
        flushMutex.release();
    }
    
    /**************************************************************************/
    /*                      Mouse event handlers                              */
    /**************************************************************************/
    
    /** Handle mouse clicks to place points.
     * 
     */
    public void mousePressed(MouseEvent e) {
        /* What we need to do depends directly on what tool is currently
         * selected.
         *
         * 0 - No tool
         * 1 - Select
         * 2 - Set Start Node
         * 3 - Add Node
         * 4 - Add Transition
         * 5 - Remove Node
         * 6 - Remove Transition
         * 7 - Move Node
         */
        
        graphNode n = getNodeUnderPoint(e.getX(), e.getY());
        
        // Check that we arent canceling the transition operation.
        if ((e.getButton() == e.BUTTON2) && inTransitionOpMode()) {
            // The user right-clicked while in transition-add mode.
            // Cancel the operation.
            cancelTransitionOp();
            return;
        }
        
        if (currentTool == 1) {
            // Select tool
            selectNode(n);
        }
        else if (currentTool == 2) {
            // Set Start Node tool
            setStartNode(n);
        }
        else if (currentTool == 3) {
            // Add Node at click position
            addNode(e.getX(), e.getY());
        }
        else if (currentTool == 4) {
            // Add Transition. Needs multiple clicks.
            // This is a multi-step operation. Click on the first node and then
            // click on the second node.  If the user clicks anywhere with
            // right-click cancel the operation.
            addTransition(n);
        }
        else if (currentTool == 5 && n != null) {
            // Remove the selected node
            selectNode(n);
            removeNode();
            setTool(0);
        }
        else if (currentTool == 6) {
            // Remove the transition. Like transition-add, this is a multi-step
            // operation.
            removeTransition(n);
        }
        else if (currentTool == 7) {
            // Ok, lets see if we can enter drag mode
            selectNode(n);
            startMove();
        }
        else {
            // If the tool is 0 we dont do anything. If it is somthing else then
            // it isnt suppoted yet.
            return;
        }
        
    }
    /** used to exit drag mode.
     */
    public void mouseReleased(MouseEvent e) {
        if (currentTool == 7) {
            // We need to stop moving if we are moving
            if (dragMode) {
                //We are dragging. Time to stop.
                stopMove();
            }
        }
    }
    /** Used for moving nodes, etc
     */
    public void mouseDragged( MouseEvent e ) {  
        if (dragMode && selectedNode != null) {
            try { moveMutex.acquire(); }
            catch (Exception ee) { ee.printStackTrace(); }
            selectedNode.position.x = e.getX() - node_radius;
            selectedNode.position.y = e.getY() - node_radius;
            moveMutex.release();
            repaint();
        }
    }
    /** Used in transition add mode to show the transition as it is being placed.
     * Allows the renderer to draw an arrow from the first node to the mouse pointer
     */
    public void mouseMoved( MouseEvent e ) {  
        if (tropMode) {
            tropPoint.setLocation(e.getX(), e.getY());
            repaint();
        }
    }  
    
    /**
     * Other stuff required by MouseListener and MouseMoveListener. They arent 
     * of any use right now.
     */
    public void mouseClicked(MouseEvent e) {

    }
    public void mouseEntered(MouseEvent e) {
    }
    public void mouseExited(MouseEvent e) {
    }
}
